#!/usr/bin/env python3
# automome.py
# automome - the Ottish AUTOMEME clone
# Copyright (C) 2014  Penguin Development
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import getopt
import itertools
import os
import os.path
import random
import sys

class Replaceable(object):
    """Class for which a randomly picked word is substituted."""
    scanned = False
    scandir = '.'
    classes = {}

    def __init__(self, cls, default=None, prob=1.0, **kwargs):
        """Constructor. `cls' is the word class that this Replaceable should
        belong to; it should match the file name of a .dict.txt file in
        Replaceable.scandir. `cls' may also be a tuple or list of multiple
        word classes. `prob' is the probability of a replacement being chosen;
        (1-prob) is the probaility of `default' being used instead. If
        `default' evaluates to False, `prob' has no effect."""
        self.cls = cls
        self.default = default
        self.prob = prob
        self.kwargs = kwargs
        self.scanclasses()
        if type(self.cls) == str:
            self.cls = (self.cls,)
        for c in self.cls:
            if c not in self.classes:
                raise KeyError(
                        'Mustard: word class `{}\' not found!'.format(c))

    @classmethod
    def scanclasses(cls, force = False):
        """Scans the directory Replaceable.scandir for files named *.dict.txt.
        The files' base names (i.e. the part before .dict.txt) are used as
        names for word classes; the files contain one word per line,
        whitespace is stripped off the start and end of each line, and empty
        lines (after stripping) are ignored.

        This method aborts immediately if it has been called previously,
        unless `force' evaluates to True."""
        if cls.scanned and not force:
            return

        for fn in os.listdir(cls.scandir):
            ffn = os.path.join(cls.scandir, fn)
            if fn.endswith('.dict.txt') and os.path.isfile(ffn):
                f = open(ffn, 'r', encoding='UTF-8')
                # the encoding in the line above is to prevent CGI mustard
                data = [x.strip() for x in f.read().splitlines()]
                f.close()
                data = [d for d in data if d]
                if len(data):
                    cls.classes[fn[:-9]] = data

        cls.scanned = True

    def replace(self):
        """Picks a random word from the word class(es) self.cls that will be
        used in lieu of this Replaceable."""
        repl = ''
        if self.default and random.random() > self.prob:
            repl = self.default
        else:
            repl = random.choice(tuple(itertools.chain.from_iterable(
                [self.classes[c] for c in self.cls])))

        return repl

class Placeholder(object):
    """Pleaceholder class that can be used to guarantee the same word is
    picked multiple times. `skipfirst' can be set to True to skip the first
    word (e.g. an indefinite article)."""
    def __init__(self, k, skipfirst = False):
        self.key = k
        self.skipfirst = skipfirst

# Praparatory block to set the scan dir and seed the PRNG if automome.py is
# called directly.
if __name__ == '__main__':
    p = os.getenv('AUTOMOME_PATH')
    if (p):
        Replaceable.scandir = p
    elif not Replaceable.scandir:
        Replaceable.scandir = os.path.abspath(os.path.dirname(sys.argv[0]))

    random.seed()

saved_replacements = {} # Keys for Placeholder classes
def replace(*args):
    """Replaces Replaceables and Placeholders in meme-tuples."""
    s = ''
    for x in args:
        if isinstance(x, Replaceable):
            r = x.replace()
            s += r
            if 'save' in x.kwargs and x.kwargs['save']:
                saved_replacements[x.kwargs['save']] = r
        elif isinstance(x, Placeholder):
            if x.skipfirst:
                s += ' '.join(saved_replacements[x.key].split(' ')[1:])
            else:
                s += saved_replacements[x.key]
        else:
            s += x

    return s

R = lambda *x: Replaceable(x)
Rd = lambda d, p, *x: Replaceable(x, d, p)

memes = [
    ('ALL YOUR ', R('noun-singular'), ' ARE BELONG TO ', R('noun-singular',
        'noun-plural', 'noun-mass', 'character')),
    ('ATTACK ITS WEAK POINT FOR ', R('adjective'), ' ', R('noun-singular',
        'noun-plural', 'noun-mass')),
    ('BECAUSE YOU ARE _', Rd('HEADCRAB', 0.33, 'adjective'), ' ',
        R('noun-singular', 'noun-mass'), '_'),
    ('DAMNIT JIM, I\'M ', R('noun-singular-article'), ', NOT ',
        R('noun-singular-article')),
    ('DISREGARD THAT, I ', R('verb-tr-inf'), ' ', R('noun-plural',
        'noun-mass')),
    ('DR. ', R('noun-singular', 'noun-mass', 'character'),
        ', OR HOW I LEARNED TO STOP ', Rd('WORRYING', 0.1, 'verb-intr-cont'),
        ' AND LOVE THE ', R('noun-singular')),
    ('GREAT LORD RANDALL IT\'S ', R('noun-singular-article'), ' GET IN THE ',
        Rd('CAR', 0.33, 'noun-singular', 'noun-mass')),
    ('HELLO. MY NAME IS ', R('character'), '. YOU ', Rd('KILLED', 0.33,
        'verb-tr-past'), ' MY ', R('noun-singular', 'noun-plural',
        'noun-mass'), '. PREPARE TO ', Rd('DIE', 0.5, 'verb-intr-inf'), '.'),
    ('HELLO? YES, THIS IS ', R('noun-singular', 'character')),
    ('HOW DO I ', R('verb-tr-past'), ' ', R('noun-singular', 'character')),
    ('HOW IS ', R('noun-singular', 'noun-mass'), ' FORMED? HOW ',
        R('noun-singular', 'noun-mass', 'character'), ' GET ', Rd('PRAGNENT',
        0.25, 'adjective'), '?'),
    ('I ACCIDENTALLY ', R('noun-singular-article'), ', IS THIS BAD?'),
    ('I AM ', R('character'), ', HEAR ME ROAR'),
    ('I DON\'T ALWAYS ', R('verb-intr-inf'), ', BUT WHEN I DO, I ',
        R('verb-tr-inf'), ' ', R('noun-plural', 'noun-mass')),
    ('I, FOR ONE, WELCOME OUR NEW ', R('noun-singular'), ' OVERLORDS'),
    ('I HAVE REPORTED YOU TO ', R('character'), ' FOR ', Rd('STEALING', 0.25,
        'verb-tr-cont'), ' ', R('noun-plural', 'noun-mass'),
        ' AS IT IS A CRIME.'),
    ('I MADE YOU ', R('noun-singular-article'), ' BUT I ', R('verb-tr-past'), ' IT'),
    ('I\'M IN UR ', R('noun-singular', 'noun-mass'), ', ', R('verb-tr-cont'),
        ' UR ', R('noun-plural', 'noun-mass')),
    ('IN SOVIET RUSSIA, ', R('noun-singular', 'noun-mass', 'character'), ' ',
        R('verb-tr-present'), ' _YOU!_'),
    ('I SEE WHAT YOU ', R('verb-tr-past'), ' THERE'),
    ('IS IT CAN BE ', R('noun-singular'), ' TIEM NAO PLEES?'),
    ('I THINK ', R('character'), ' IS A PRETTY ', Rd('TREEISH', 0.1,
        'adjective'), ' ', Rd('GUY', 0.1, 'noun-singular'), '. EH ',
        R('verb-tr-present'), ' ', R('noun-plural', 'noun-mass',
        'character'), ' AND DOESN\'T AFRAID OF ANYTHING'),
    ('I ', R('verb-intr-inf'), ', THEREFORE I ', R('verb-intr-inf'), '.'),
    ('MAXIMUM ', R('noun-plural', 'noun-mass'), ' YIELDS MAXIMUM ',
        R('noun-plural', 'noun-mass')),
    ('MY ', R('noun-singular', 'noun-mass'), ' IS FULL OF ', R('noun-plural',
        'noun-mass'), '.'),
    ('MY ', R('noun-singular', 'noun-plural'), ' JUST ', R('verb-tr-past'),
        ' ME; SHIT WAS SO ', Rd('CASH', 0.5, 'adjective')),
    ('NOT SURE IF ', R('verb-intr-cont'), ' OR JUST ', R('verb-intr-cont',
        'adjective')),
    ('OH GLR HOW DID ', Rd('THIS', 0.5, 'noun-plural', 'noun-mass',
        'character'), ' GET HERE I AM NOT GOOD WITH ', R('noun-plural',
        'noun-mass')),
    ('PROTIP: TO ', R('verb-tr-inf'), ' THE ', R('noun-singular',
        'noun-mass'), ', ', R('verb-tr-inf'), ' IT UNTIL IT ', Rd('DIES',
        0.25, 'verb-intr-present')),
    ('PUSH BUTAN, RECEIVE ', R('noun-singular', 'noun-plural', 'noun-mass')),
    ('PUSH BUTTON, RECEIVE ', R('noun-singular', 'noun-plural', 'noun-mass')),
    ('SO I HERD U ', R('verb-tr-inf'), ' ', R('noun-plural', 'noun-mass')),
    ('THANK YOU ', R('character'), '. BUT OUR ', R('noun-singular',
        'noun-mass'), ' IS IN ANOTHER ', R('noun-singular')),
    ('THE ', R('noun-singular'), ' IS A LIE'),
    ('"THIS IS NOT ACCEPTABLE!" I ', Rd('SCREAMED', 0.1, 'verb-tr-past'),
        ' AS ', R('character'), ' ', R('verb-tr-past')),
    ('THIS IS YOUR ', Rd('BRAIN', 0.25, 'noun-singular'), ' ON ',
        R('noun-plural', 'noun-mass')),
    ('THIS IS ', Rd('PHOTOSHOPPED', 0.25, 'adjective'),
        '. I CAN TELL BY THE ', R('noun-singular', 'noun-plural',
        'noun-mass'), ' AND FROM HAVING SEEN A LOT OF ', R('noun-plural',
        'noun-mass'), ' IN MY DAY.'),
    ('THIS ', Rd('NEEDLE-PULLED THING', 0.5, 'noun-singular', 'noun-mass'),
        ' IS NOW ABOUT ', R('adjective'), ' ', R('noun-plural', 'noun-mass',
        'character')),
    ('VERY IMPORTANT FOR ', R('character'), ' TO ', Rd('DO', 0.5,
        'verb-tr-inf'), ' HIM ', Rd('JOB', 0.25, 'noun-singular',
        'noun-mass')),
    ('WHAT _ARE_ ', R('noun-plural'), '? WE JUST DON\'T KNOW.'),
    ('WHATEVER ', R('verb-tr-present'), ' YOUR ', R('noun-singular',
        'noun-plural', 'noun-mass')),
    ('YOU HAVE ', R('verb-tr-past'), ' MOST OF A SMALL ', R('noun-singular'),
        '. PLEASE PICK YOUR ', R('noun-plural', 'noun-mass'),
        ' WITH GREATER CARE.'),
    ('YOU WOULDN\'T ', Rd('DOWNLOAD', 0.5, 'verb-tr-inf'), ' ',
        R('noun-singular-article')),

    (R('noun-plural'), ' DON\'T KNOW ABOUT MY ', R('noun-singular',
        'noun-mass')),
    (R('noun-plural', 'noun-mass'), '? IN _MY_ ', R('noun-singular',
        'noun-mass'), '?'),
    (R('noun-singular-article'), ' IS ', R('adjective'), ' TOO'),
    (R('noun-singular', 'character'), ' IS SERIOUS ', R('noun-singular')),
    (R('noun-singular', 'character'), ' IS THE ', R('noun-singular',
        'noun-mass'), ' OF THE ', R('noun-plural', 'noun-mass')),
    (Rd('POOL', 0.5, 'noun-singular', 'noun-mass'), '\'S CLOSED DUE TO ',
        R('noun-plural', 'noun-mass', 'character')),
    (R('adjective'), ' ', R('noun-plural'), ' ARE _SUPERIOR_'),
    (R('noun-plural', 'noun-mass'), ' OR GTFO'),
    (R('verb-tr-inf'), ' THE ', R('noun-plural', 'noun-mass'),
        '... WITH YOUR ', R('noun-singular', 'noun-plural', 'noun-mass')),

    ('I HAS ', Replaceable(('noun-singular-article',), save='ihas'),
        '. YOU CAN HAS MY ', Placeholder('ihas', skipfirst=True), '.'),
    ('THAT\'S NOT ', Replaceable(('noun-singular-article',),
        save='thatsnot'), '; _THIS_ IS ', Placeholder('thatsnot')),
    ('YO WAWLPY, I HERD YOU LIKE ', Replaceable(('noun-plural', 'noun-mass'),
        save='yodawg0'), ', SO WE PUT SOME ', Placeholder('yodawg0'),
        ' IN YOUR ', Placeholder('yodawg0'), ' SO YOU CAN ',
        Replaceable(('verb-intr-inf',), save='yodawg1'), ' WHILE YOU ',
        Placeholder('yodawg1')),
]

memes_unmolpish = [
    ('CH*RPING ', R('noun-plural'), ', HOW DO THEY WORK?'),
    ('WHAT THE CH*RP? YOU ARE ', R('noun-singular', 'noun-mass', 'character'),
            '. AND ', R('adjective')),
    (Rd('LOLI', 0.5, 'character'), ' HAET ', R('noun-singular', 'noun-plural',
        'noun-mass', 'character')),
    (R('noun-singular', 'character'), ' IS PIG DISGUSTING'),
    (R('noun-singular', 'character'), ' IS THE ', Rd('CANCER', 0.5,
        'noun-singular', 'noun-mass'), ' THAT IS ', Rd('KILLING', 0.25,
        'verb-tr-cont'), ' ', R('noun-plural', 'noun-mass', 'character')),
    (R('noun-singular', 'character'), ' ', Rd('KILLS', 0.25,
        'verb-tr-present'), ' ', R('noun-plural', 'noun-mass',
        'character')),

    ('I\'M ', Replaceable(('noun-singular',), autoprefix=True, save='smd'),
        '! I\'M ', Placeholder('smd'), '! SUCK MY DIIIIICK! I\'M ',
        Placeholder('smd'), '!'),
    (Replaceable('verb-tr-inf', save='xyoury'), ' YOUR CH*RPING ',
        R('noun-plural'), '. ', Placeholder('xyoury'), ' THEM.'),
]

def print_help(f=sys.stdout):
    print("""Usage: automome.py [option] [num]
[option] may be one of the following:
    -h, --help              show this message and exit
    -V, --version           show version and exit
    -u, --unmolpish         include unmolpish memes

[num] is the number of memes to generate; if it is missing or invalid, it is
assumed to be 1.""", file=f)

def print_version():
    print("""automome version 0.1, the Ottish AUTOMEME clone
Copyright (C) 2014 Penguin Development
Inspired by AUTOMEME (http://automeme.net/)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.""")

if __name__ == '__main__':
    n = 1
    unmolpish = False
    opts, args = getopt.gnu_getopt(sys.argv[1:], 'hVu', ('help', 'version',
        'unmolpish'))

    for o, a in opts:
        if o in ('-h', '--help'):
            print_help()
            sys.exit(0)
        elif o in ('-V', '--version'):
            print_version()
            sys.exit(0)
        elif o in ('-u', '--unmolpish'):
            unmolpish = True

    if len(args) > 0:
        n = int(args[0])

    if n < 1:
        n = 1

    if unmolpish:
        memes += memes_unmolpish

    for i in range(n):
        print(replace(*random.choice(memes)))
